# -*- coding: utf-8 -*-
"""
.. module:: decoder_softmax

"""

import torch
from torch import nn

from .utils import sequence_mask


class DecoderSoftmax(nn.Module):
    """
    RNN + CRF
    """

    def __init__(self):
        """init"""
        super(DecoderSoftmax, self).__init__()
        self._cross_entropy_loss = nn.CrossEntropyLoss(
            ignore_index=0, reduction='sum')

    def compute_loss(self, encoder_output, tags, lengths):
        """Softmax loss"""
        batch_size, seq_len = encoder_output.size()[:2]
        seq_len = encoder_output.size(1)
        masks = sequence_mask(lengths, seq_len)
        encoder_output = encoder_output * masks.view(batch_size, seq_len, 1)
        tags = tags * masks.long()
        encoder_output = encoder_output.view(batch_size * seq_len, -1)
        tags = tags.view(-1)
        return self._cross_entropy_loss(encoder_output, tags)

    def forward(self, encoder_output):
        """Softmax Result"""
        scores = encoder_output.max(dim=2)
        _, pred = torch.max(encoder_output, dim=2)
        return scores, pred.cpu().detach().numpy()



# -*- coding: utf-8 -*-
"""
.. module:: encoder_rnn

"""

import torch
from torch import nn


class EncoderRNN(nn.Module):
    """
    RNN Encoder
    """

    def __init__(self, vocab_size, tag_to_ix, embedding_dim, hidden_dim,
                 num_layers, bidirectional, embedding_dropout_p, rnn_dropout_p,
                 rnn_type, device, embedding_trainable, use_char,
                 char_hidden_dim):
        """
        Word Embedding and RNN
        """
        super(EncoderRNN, self).__init__()

        if rnn_type not in ('lstm', 'gru'):
            raise Exception('Invalid rnn_type')

        self._vocab_size = vocab_size
        self._tag_to_ix = tag_to_ix
        self._embedding_dim = embedding_dim
        self._hidden_dim = hidden_dim
        self._num_layers = num_layers
        self._bidirectional = bidirectional
        self._embedding_dropout_p = embedding_dropout_p
        self._rnn_dropout_p = rnn_dropout_p
        self._rnn_type = rnn_type
        self._device = device
        self._embedding_trainable = embedding_trainable
        self._use_char = use_char
        self._char_hidden_dim = char_hidden_dim

        self._tagset_size = len(tag_to_ix)
        self._directions = 2 if bidirectional else 1
        self._embedding_dropout = nn.Dropout(p=self._embedding_dropout_p)
        self._rnn_dropout = nn.Dropout(p=self._rnn_dropout_p)

        # Embedding Layer
        self._word_embeds = nn.Embedding(
            vocab_size, embedding_dim, padding_idx=0)
        self._word_embeds.weight.requires_grad = self._embedding_trainable

        input_size = embedding_dim
        if use_char:
            input_size += char_hidden_dim

        # RNN Layer
        if rnn_type == 'lstm':
            self._rnn = nn.LSTM(
                input_size,
                hidden_dim // self._directions,
                num_layers=num_layers,
                bidirectional=bidirectional,
                dropout=rnn_dropout_p if num_layers > 1 else 0)
        elif rnn_type == 'gru':
            self._rnn = nn.GRU(
                input_size,
                hidden_dim // self._directions,
                num_layers=num_layers,
                bidirectional=bidirectional,
                dropout=rnn_dropout_p if num_layers > 1 else 0)

        # Tag projection
        self._hidden2tag = nn.Linear(hidden_dim, self._tagset_size)

    def load_embedding(self, pretrained_embedding):
        """Load other pre-trained embedding vectors"""
        word_embeds = self._word_embeds
        assert len(pretrained_embedding.shape) == 2
        assert pretrained_embedding.shape[0] == word_embeds.weight.size(0)
        assert pretrained_embedding.shape[1] == word_embeds.weight.size(1)
        print('loaded pre-trained vectors', pretrained_embedding.shape)
        pretrained_embedding = torch.from_numpy(pretrained_embedding)
        word_embeds.weight.data.copy_(pretrained_embedding)
        word_embeds.weight.requires_grad = self._embedding_trainable

    def forward(self, sentence, lengths, char_feats=None):
        """Get the output of RNN model"""
        # sentence dim: [batch_size, seq_len]
        batch_size, seq_len = sentence.size()
        # embeds dim: [batch_size, seq_len, embedding_size]
        embeds = self._word_embeds(sentence)
        if char_feats is not None:
            embeds = torch.cat([embeds, char_feats], 2)
        embeds = self._embedding_dropout(embeds)
        # embeds dim: [seq_len, batch_size, embedding_size]
        embeds = embeds.transpose(1, 0)

        packed = nn.utils.rnn.pack_padded_sequence(
            embeds, lengths, batch_first=False)
        rnn_out, _ = self._rnn(packed)
        # rnn_out dim: [batch_size, seq_len, hidden_dim]
        rnn_out, _ = nn.utils.rnn.pad_packed_sequence(
            rnn_out, batch_first=True)
        rnn_out = self._rnn_dropout(rnn_out)
        # rnn_feats dim: [batch_size, seq_len, target_dim]
        rnn_feats = self._hidden2tag(rnn_out)
        return rnn_feats.view(batch_size, seq_len, -1)



"""
.. module:: char_rnn

Char + RNN Features

"""

import torch
from torch import nn


class CharRNN(nn.Module):
    """
    Char + RNN
    """

    def __init__(self, char_vocab_size, char_max_len, char_embedding_dim,
                 char_hidden_dim, char_dropout_p, char_bidirectional, rnn_type,
                 device, embedding_trainable):
        """init"""
        super(CharRNN, self).__init__()
        self._char_vocab_size = char_vocab_size
        self._char_max_len = char_max_len
        self._char_embedding_dim = char_embedding_dim
        self._char_hidden_dim = char_hidden_dim
        self._char_dropout_p = char_dropout_p
        self._char_bidirectional = char_bidirectional
        self._rnn_type = rnn_type
        self._device = device
        self._embedding_trainable = embedding_trainable

        self._char_directions = 2 if char_bidirectional else 1
        self._char_dropout = nn.Dropout(self._char_dropout_p)
        self._char_embeds = nn.Embedding(
            char_vocab_size, char_embedding_dim, padding_idx=0)
        self._char_embeds.weight.requires_grad = self._embedding_trainable

        if rnn_type == 'lstm':
            self._char_rnn = nn.LSTM(
                char_embedding_dim,
                char_hidden_dim // self._char_directions,
                num_layers=1,
                bidirectional=char_bidirectional)
        elif rnn_type == 'gru':
            self._char_rnn = nn.GRU(
                char_embedding_dim,
                char_hidden_dim // self._char_directions,
                num_layers=1,
                bidirectional=char_bidirectional)

    def forward(self, chars, charlens):
        assert len(chars.size()) == 3
        batch_size = chars.size(0)
        seq_len = chars.size(1)
        char_len = chars.size(2)
        total_word = batch_size * seq_len
        chars = chars.view(total_word, char_len)
        chars_lengths = charlens.view(total_word)
        chars_lengths, perm_index = chars_lengths.sort(0, descending=True)

        # char_feats dim: [total_word, char_len, char_embedding_size]
        char_feats = self._char_embeds(chars)
        # dropout char embedding
        char_feats = self._char_dropout(char_feats)
        char_feats = char_feats[perm_index]
        char_feats = char_feats[chars_lengths > 0]
        chars_lengths = chars_lengths[chars_lengths > 0]

        char_feats = char_feats.transpose(1, 0)
        char_feats = nn.utils.rnn.pack_padded_sequence(
            char_feats, chars_lengths, batch_first=False)

        # Get last hidden as features
        char_feats, char_hidden = self._char_rnn(char_feats)
        # char_hidden dim: [directions, total_word - padding, hidden_dim]
        if isinstance(char_hidden, tuple):
            char_hidden = char_hidden[0]

        # char_hidden dim: [total_word - padding, directions, hidden_dim]
        char_hidden = char_hidden.transpose(1, 0)
        # char_hidden dim: [total_word - padding, directions * hidden_dim]
        char_hidden = char_hidden.contiguous().view(char_hidden.size(0), -1)

        char_feats, _ = nn.utils.rnn.pad_packed_sequence(
            char_feats, batch_first=True, total_length=self._char_max_len)

        char_feats = char_hidden
        pad_length = total_word - char_feats.size(0)
        pad = torch.zeros((pad_length, char_feats.size(1)),
                          device=self._device)
        char_feats = torch.cat([char_feats, pad], dim=0)

        char_feats[perm_index] = char_feats[torch.arange(0, len(char_feats))]

        char_feats = char_feats.contiguous().view(batch_size, seq_len,
                                                  self._char_hidden_dim)
        return char_feats



# -*- coding: utf-8 -*-
"""
.. module:: decoder_crf

"""

import numpy as np

import torch
from torch import nn

from .utils import START_TAG, STOP_TAG, sequence_mask

# Transfer disable
NO_TRANS = -10000.


class DecoderCRF(nn.Module):
    """
    decoder + CRF
    """

    def __init__(self, tag_to_ix, device):
        """init"""
        super(DecoderCRF, self).__init__()
        self._tag_to_ix = tag_to_ix
        self._device = device
        self._stop_idx = self._tag_to_ix[STOP_TAG]
        self._start_idx = self._tag_to_ix[START_TAG]
        self._tagset_size = len(tag_to_ix)
        self._transitions = nn.Parameter(
            torch.zeros(self._tagset_size, self._tagset_size))
        # These two statements enforce the constraint that we never transfer
        # to the start tag and we never transfer from the stop tag
        nn.init.constant_(self._transitions[self._start_idx, :], NO_TRANS)
        nn.init.constant_(self._transitions.data[:, self._stop_idx], NO_TRANS)
        # PAD_TAG
        nn.init.constant_(self._transitions.data[:, 0], NO_TRANS)
        nn.init.constant_(self._transitions.data[0, :], NO_TRANS)

    def _forward_alg_single(self, feats):
        # Do the forward algorithm to compute the partition function
        init_alphas = torch.full((1, self._tagset_size),
                                 NO_TRANS,
                                 device=self._device)
        # START_TAG has all of the score.
        init_alphas[0][self._start_idx] = 0.

        # Wrap in a variable so that we will get automatic backprop
        forward_var = init_alphas

        # Iterate through the sentence
        for feat in feats:
            alphas_t = []  # The forward tensors at this timestep
            for next_tag in range(self._tagset_size):
                # broadcast the emission score: it is the same regardless of
                # the previous tag
                emit_score = feat[next_tag].view(1, -1).expand(
                    1, self._tagset_size)
                # the ith entry of trans_score is the score of transitioning to
                # next_tag from i
                trans_score = self._transitions[next_tag].view(1, -1)
                # The ith entry of next_tag_var is the value for the
                # edge (i -> next_tag) before we do log-sum-exp
                next_tag_var = forward_var + trans_score + emit_score
                # The forward variable for this tag is log-sum-exp of all the
                # scores.
                alphas_t.append(
                    torch.logsumexp(next_tag_var.view(1, -1), 1).view(1))
            forward_var = torch.cat(alphas_t).view(1, -1)
        terminal_var = forward_var + self._transitions[self._stop_idx]
        alpha = torch.logsumexp(terminal_var.view(1, -1), 1)
        return alpha

    def _forward_alg(self, feats_batch, lengths_batch):
        """Get CRF result"""
        # feats_batch dim: [batch_size, seq_len, target_dim]
        # lengths_batch dim: [batch_size]
        batch_size = feats_batch.size(0)
        alpha_batch = []

        for i in range(batch_size):
            length = lengths_batch[i]
            feats = feats_batch[i][:length]
            alpha = self._forward_alg_single(feats)
            alpha_batch.append(alpha)
        return torch.stack(alpha_batch).view(-1)

    def _forward_alg_batch(self, feats, lengths):
        """Get CRF result, batch version"""
        # feats dim: [batch_size, seq_len, target_dim]
        # lengths dim: [batch_size]
        # masks dim: [batch_size, seq_len]
        batch_size, seq_len = feats.size()[:2]

        # Do the forward algorithm to compute the partition function
        init_alphas = torch.full((batch_size, 1, self._tagset_size),
                                 NO_TRANS,
                                 requires_grad=False,
                                 device=self._device)
        # START_TAG has all of the score.
        init_alphas[:, 0, self._start_idx] = 0.

        # Wrap in a variable so that we will get automatic backprop
        forward_var = init_alphas

        # feats dim: [seq_len, batch_size, target_dim]
        feats = feats.transpose(1, 0)
        feats = feats.view(seq_len, batch_size, self._tagset_size, 1)

        trans_score = self._transitions.view(1, self._tagset_size,
                                             self._tagset_size)

        forward_var_batch = []
        for emit_score in feats:
            forward_var = forward_var.view(batch_size, 1, self._tagset_size)
            next_tag_var = forward_var + \
                trans_score + \
                emit_score
            next_tag_var = torch.logsumexp(next_tag_var, dim=2)
            forward_var = next_tag_var
            forward_var_batch.append(forward_var)

        forward_var_batch = torch.stack(forward_var_batch)
        forward_var = forward_var_batch.gather(
            0, (lengths.view(1, batch_size, 1) - 1).expand(
                1, batch_size, self._tagset_size)).squeeze()
        forward_var = forward_var.view(batch_size, -1)

        alpha = forward_var + self._transitions[self._stop_idx]
        alpha = torch.logsumexp(alpha, 1)
        return alpha.view(-1)

    def _score_sentence_single(self, feats, tags):
        # Gives the score of a provided tag sequence
        score = torch.zeros(1, device=self._device)
        tags = torch.cat([
            torch.tensor([self._start_idx],
                         dtype=torch.long,
                         device=self._device), tags
        ])
        for i, feat in enumerate(feats):
            score = score + \
                self._transitions[tags[i + 1], tags[i]] + feat[tags[i + 1]]
        score = score + self._transitions[self._stop_idx, tags[-1]]
        return score

    def _score_sentence(self, feats_batch, tags_batch, lengths_batch):
        """Get the gold standard of sentences
        Gives the score of a provided tag sequence
        feats_batch dim: [batch_size, seq_len, target_dim]
        tags_batch dim: [batch_size, seq_len]
        lengths_batch dim: [batch_size]
        """
        batch_size = feats_batch.size(0)
        score_batch = []
        for i in range(batch_size):
            score = torch.zeros(1, device=self._device)
            length = lengths_batch[i]
            feats = feats_batch[i][:length]
            tags = tags_batch[i][:length]
            score = self._score_sentence_single(feats, tags)
            score_batch.append(score)
        score = torch.stack(score_batch).view(-1)
        return score.view(-1)

    def _score_sentence_batch(self, feats, tags, lengths, masks):
        """Get the gold standard of sentences, batch version
        Gives the score of a provided tag sequence
        feats dim: [batch_size, seq_len, target_dim]
        tags dim: [batch_size, seq_len]
        lengths dim: [batch_size]
        masks dim: [batch_size, seq_len]
        """
        batch_size = feats.size(0)
        score = torch.zeros(
            batch_size, requires_grad=True, device=self._device)
        start = torch.tensor([self._start_idx],
                             dtype=torch.long,
                             device=self._device).expand(batch_size, 1)
        tags = torch.cat([start, tags], dim=1)
        # feats dim: [seq_len, batch_size, hidden_dim]
        feats = feats.transpose(1, 0)
        for j, feat in enumerate(feats):
            mask = masks[:, j].contiguous().view(batch_size, 1)
            tag_next, tag_cur = tags[:, j + 1], tags[:, j]
            score = score \
                + self._transitions[tag_next, tag_cur] * mask \
                + feat[:, tag_next] * mask
        ends = tags.gather(1, lengths.view(-1, 1))
        trans = self._transitions[self._stop_idx].expand(
            batch_size, self._tagset_size).gather(1, ends.view(-1, 1))
        score = score + trans.view(-1)
        score = score.diag()
        return score.view(-1)

    def _viterbi_decode_single(self, feats):
        backpointers = []

        # Initialize the viterbi variables in log space
        init_vvars = torch.full((1, self._tagset_size),
                                NO_TRANS,
                                device=self._device)
        init_vvars[0][self._start_idx] = 0

        # forward_var at step i holds the viterbi variables for step i-1
        forward_var = init_vvars
        for feat in feats:
            bptrs_t = []  # holds the backpointers for this step
            viterbivars_t = []  # holds the viterbi variables for this step

            for next_tag in range(self._tagset_size):
                # next_tag_var[i] holds the viterbi variable for tag i at the
                # previous step, plus the score of transitioning
                # from tag i to next_tag.
                # We don't include the emission scores here because the max
                # does not depend on them (we add them in below)
                next_tag_var = forward_var + self._transitions[next_tag]
                best_tag_id = torch.argmax(next_tag_var, 1)[0]
                bptrs_t.append(best_tag_id)
                viterbivars_t.append(next_tag_var[0][best_tag_id].view(1))
            # Now add in the emission scores, and assign forward_var to the set
            # of viterbi variables we just computed
            forward_var = (torch.cat(viterbivars_t) + feat).view(1, -1)
            backpointers.append(bptrs_t)

        # Transition to STOP_TAG
        terminal_var = forward_var + \
            self._transitions[self._stop_idx]
        best_tag_id = torch.argmax(terminal_var, 1)[0]
        path_score = terminal_var[0][best_tag_id]

        # Follow the back pointers to decode the best path.
        best_path = [best_tag_id]
        for bptrs_t in reversed(backpointers):
            best_tag_id = bptrs_t[best_tag_id]
            best_path.append(best_tag_id)
        # Pop off the start tag (we dont want to return that to the caller)
        start = best_path.pop()
        assert start == self._start_idx  # Sanity check
        best_path.reverse()
        return path_score, best_path

    def _viterbi_decode(self, feats_batch, lengths):
        """Inference result
        feats_batch dim: [batch_size, seq_len, target_dim]
        lengths dim: [batch_size]
        """
        batch_size = feats_batch.size(0)
        path_scores, best_paths = [], []
        for i in range(batch_size):
            length = lengths[i]
            feats = feats_batch[i][:length]
            path_score, best_path = self._viterbi_decode_single(feats)
            path_scores.append(path_score)
            best_paths.append(best_path)
        path_scores = torch.stack(path_scores)
        best_paths = np.asarray([np.asarray(x) for x in best_paths])
        return path_scores, best_paths

    def _viterbi_decode_batch(self, feats, lengths, masks):
        """Inference result"""
        # feats dim: [batch_size, seq_len, target_dim]
        batch_size = feats.size(0)

        backpointers = []

        # Initialize the viterbi variables in log space
        # init_vvars dim: [batch_size, tagset_size]
        init_vvars = torch.full((batch_size, self._tagset_size),
                                NO_TRANS,
                                device=self._device)
        init_vvars[:, self._start_idx] = 0

        # forward_var at step i holds the viterbi variables for step i-1
        forward_var = init_vvars

        # feats dim: [seq_len, batch_size, target_dim]
        feats = feats.transpose(1, 0)
        # feat dim: [batch_size, target_dim]
        bptrs_t = None
        for i, feat in enumerate(feats):
            mask = masks[:, i]
            # bptrs_t_old = bptrs_
            bptrs_t = []  # holds the backpointers for this step
            viterbivars_t = []  # holds the viterbi variables for this step

            for next_tag in range(self._tagset_size):
                # next_tag_var[i] holds the viterbi variable for tag i at the
                # previous step, plus the score of transitioning
                # from tag i to next_tag.
                # We don't include the emission scores here because the max
                # does not depend on them (we add them in below)
                # next_tag_var dim: [batch_size, tagset_size]
                next_tag_var = forward_var + self._transitions[next_tag].view(
                    1, -1).expand(batch_size, self._tagset_size)
                best_tag_id = torch.argmax(next_tag_var, 1)
                bptrs_t.append(best_tag_id.view(1, -1))
                tmp = next_tag_var.gather(1, best_tag_id.view(-1, 1)).squeeze()
                viterbivars_t.append(tmp.view(1, batch_size))
            # Now add in the emission scores, and assign forward_var to the set
            # of viterbi variables we just computed
            # forward_var dim: [batch_size, tagset_size]
            tmp = torch.cat(viterbivars_t)
            tmp = tmp.permute(1, 0)
            tmp_new_forward_var = (tmp + feat).view(batch_size, -1)

            mask_table = mask.view(-1, 1).expand(-1, self._tagset_size)
            forward_var = tmp_new_forward_var * mask_table + forward_var * (
                mask_table - 1).abs()

            bptrs_t = torch.cat(bptrs_t)

            backpointers.append(bptrs_t)

        # Transition to STOP_TAG
        # terminal_var dim: [1, tagset_size]
        terminal_var = forward_var + \
            self._transitions[self._stop_idx]
        best_tag_id = torch.argmax(terminal_var, 1)
        path_score = [terminal_var[i][x] for i, x in enumerate(best_tag_id)]

        # Follow the back pointers to decode the best path.
        backpointers = torch.stack(backpointers)
        # backpointers dim: [seq_len, batch_size, tagset_size]
        backpointers = backpointers.permute(0, 2, 1)
        # print('backpointers', backpointers.size())
        best_path = [best_tag_id]
        for bptrs_t in reversed(backpointers):
            if batch_size > 1:
                best_tag_id = bptrs_t.gather(1, best_tag_id.view(-1,
                                                                 1)).squeeze()
                best_path.append(best_tag_id)
            else:
                best_tag_id = bptrs_t[0][best_tag_id]
                best_path.append(best_tag_id)
        # Pop off the start tag (we dont want to return that to the caller)
        # best_path.pop()
        # check start
        start = best_path.pop()
        assert np.sum(start.cpu().detach().numpy() -
                      np.array([self._start_idx] * batch_size)) <= 1e-5
        best_path.reverse()

        path_score = torch.stack(path_score)
        best_path = torch.stack(best_path)
        best_path = best_path.permute(1, 0).cpu().detach().numpy()
        best_path = np.asarray(
            [np.asarray(x[:i]) for i, x in zip(lengths, best_path)])

        return path_score, best_path

    def compute_loss(self, encoder_output, tags, lengths):
        """Calculate loss
        encoder_output dim: [batch_size, seq_len, embedding_dim]
        tags dim: [batch_size, seq_len]
        lengths dim: [batch_size]
        """

        seq_len = encoder_output.size(1)
        masks = sequence_mask(lengths, seq_len)

        forward_score = self._forward_alg_batch(encoder_output, lengths)
        gold_score = self._score_sentence_batch(encoder_output, tags, lengths,
                                                masks)

        loss = torch.sum(forward_score) - torch.sum(gold_score)
        # loss = loss / batch_size
        return loss

    def forward(self, encoder_output, lengths, masks):
        return self._viterbi_decode_batch(encoder_output, lengths, masks)



"""
.. module:: char_cnn

Char + CNN Features

Basically copy from:
https://github.com/jiesutd/NCRFpp/blob/master/model/charbilstm.py

"""

import torch
import torch.nn.functional as F
from torch import nn


class CharCNN(torch.nn.Module):
    """Char + CNN
    """

    def __init__(self, char_vocab_size, char_max_len, char_embedding_dim,
                 char_hidden_dim, char_dropout_p, device, embedding_trainable):
        """init
        """
        super(CharCNN, self).__init__()
        self._char_vocab_size = char_vocab_size
        self._char_max_len = char_max_len
        self._char_embedding_dim = char_embedding_dim
        self._char_hidden_dim = char_hidden_dim
        self._char_dropout_p = char_dropout_p
        self._device = device
        self._embedding_trainable = embedding_trainable

        self._char_cnn = nn.Conv1d(
            char_embedding_dim, char_hidden_dim, kernel_size=3, padding=1)

        self._char_dropout = nn.Dropout(self._char_dropout_p)
        self._char_embeds = nn.Embedding(
            char_vocab_size, char_embedding_dim, padding_idx=0)
        self._char_embeds.weight.requires_grad = self._embedding_trainable

    def forward(self, chars, charlens=None):
        assert len(chars.size()) == 3
        batch_size = chars.size(0)
        seq_len = chars.size(1)
        total_word = batch_size * seq_len
        chars = chars.view(total_word, self._char_max_len)
        char_feats = self._char_embeds(chars)
        char_feats = self._char_dropout(char_feats)
        char_feats = char_feats.transpose(2, 1)
        char_feats = self._char_cnn(char_feats)
        char_feats = F.max_pool1d(char_feats, char_feats.size(2))
        char_feats = char_feats.view(batch_size, seq_len,
                                     self._char_hidden_dim)

        return char_feats


def test():
    """Test CharCNN"""
    import torch

    char_vocab_size = 10
    char_max_len = 10
    char_embedding_dim = 30
    char_hidden_dim = 50
    char_dropout_p = 0.5
    device = torch.device('cpu')
    embedding_trainable = True

    ccnn = CharCNN(char_vocab_size, char_max_len, char_embedding_dim,
                   char_hidden_dim, char_dropout_p, device,
                   embedding_trainable)

    batch_input = torch.rand(10, 10, 10) * 10
    batch_input = batch_input.long()
    out = ccnn(batch_input)
    print(out.size())


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
"""Test TaggerModel"""

import numpy as np

import torch
from torch_tagger.tagger_model import TaggerModel
from torch_tagger.utils import START_TAG, STOP_TAG, sequence_mask


def same(amat, bmat):
    """Test two matrix is same"""
    return np.sum((amat.flatten() - bmat.flatten())**2)


def test_model(batch_size=2,
               seq_len=9,
               vocab_size=10,
               embedding_dim=10,
               hidden_dim=10):
    """Test model entry"""
    device = torch.device('cpu')
    tag_to_ix = {
        'a': 0,
        'b': 1,
        START_TAG: 2,
        STOP_TAG: 3,
    }

    scale = 1000.

    model = TaggerModel(
        vocab_size,
        tag_to_ix,
        embedding_dim,
        hidden_dim,
        num_layers=1,
        bidirectional=True,
        embedding_dropout_p=0,
        rnn_dropout_p=0,
        rnn_type='lstm',
        device=device,
        embedding_trainable=True,
        use_crf=True,
        use_char='lstm',
        char_vocab_size=10,
        char_max_len=10,
        char_embedding_dim=10,
        char_hidden_dim=10,
        char_dropout_p=0,
        char_bidirectional=True,
        average_loss=False)
    # feats_batch dim: [batch_size, seq_len, target_dim]
    # lengths_batch dim: [batch_size]
    target_dim = len(tag_to_ix)
    feats = torch.randn(batch_size, seq_len, target_dim) * scale

    # sentences = torch.randint(0, vocab_size, (batch_size, seq_len)).long()
    tags = torch.randint(0, target_dim, (batch_size, seq_len)).long()
    lengths = torch.tensor(
        # [seq_len] * batch_size
        np.array(
            sorted(
                np.random.randint(1, seq_len, size=batch_size).tolist(),
                reverse=True))).view(-1)
    masks = sequence_mask(lengths, seq_len)

    decode_input = torch.rand(batch_size, seq_len, len(tag_to_ix))
    score, pred = model._decoder._viterbi_decode(decode_input, lengths)
    score_b, pred_b = model._decoder._viterbi_decode_batch(
        decode_input, lengths, masks)
    score = score.cpu().detach().numpy()
    score_b = score_b.cpu().detach().numpy()
    assert score.shape == score_b.shape
    assert np.sum(np.sum((score - score_b)**2)) < 1e-10
    assert pred.shape == pred_b.shape, 'batch_size: {} {} {}'.format(
        batch_size, pred, pred_b)

    diff = [np.sum(a - b) for a, b in zip(pred, pred_b)]

    assert np.sum(np.sum(diff)) < 1e-10, 'batch_size: {} {} {}'.format(
        batch_size, pred, pred_b)

    feats = feats * masks.view(batch_size, seq_len, 1)
    ret = model._decoder._forward_alg(feats, lengths)
    retb = model._decoder._forward_alg_batch(feats, lengths)
    ret, retb = ret.cpu().detach().numpy(), retb.cpu().detach().numpy()
    diff = same(ret, retb)
    # print('forward diff', diff, ret[:3])
    assert diff < 1e-10, 'diff: {}, {}, {}'.format(diff, ret, retb)

    rets = model._decoder._score_sentence(feats, tags, lengths)
    retsb = model._decoder._score_sentence_batch(feats, tags, lengths, masks)
    diff = same(rets.cpu().detach().numpy(), retsb.cpu().detach().numpy())
    # print('score diff', diff, rets[:3])
    assert diff < 1e-10, 'diff: {}'.format(diff)


def test_models():
    """Test different parameters"""
    temp = '''batch_size {} seq_len {} ''' + \
        '''vocab_size {} embedding_dim {} hidden_dim {}'''
    params = []
    for batch_size in (1, 3, 9):
        for seq_len in (2, 3, 9, 100):
            for vocab_size in (3, 9, 100):
                for embedding_dim in (10, 100, 500):
                    for hidden_dim in (10, 100, 500):
                        params.append((batch_size, seq_len, vocab_size,
                                       embedding_dim, hidden_dim))
    for param in params:
        try:
            test_model(*param)
        except AssertionError as err:
            print(
                temp.format(batch_size, seq_len, vocab_size, embedding_dim,
                            hidden_dim))
            raise err


if __name__ == '__main__':
    test_models()



# -*- coding: utf-8 -*-
"""Test Entry"""

from tests.grid import test_grid
from tests.model import test_model
from tests.tagger import test_tagger
from tests.utils import test_utils

if __name__ == '__main__':
    test_utils()
    test_model()
    test_tagger()
    test_grid()



#!/usr/bin/env python3
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: InfinityFuture
# Mail: infinityfuture@foxmail.com
# Created Time: 2018-09-06 10:00:00
#############################################

import os
from setuptools import setup, find_packages

ON_RTD = os.environ.get('READTHEDOCS') == 'True'
if not ON_RTD:
    INSTALL_REQUIRES = [
        'torch', 'tqdm', 'scikit-learn', 'numpy', 'scipy', 'pypinyin'
    ]
else:
    INSTALL_REQUIRES = []

VERSION = os.path.join(
    os.path.realpath(os.path.dirname(__file__)),
    'version.txt'
)

setup(
    name='torch-tagger',
    version=open(VERSION, 'r').read().strip(),
    keywords=('pip', 'pytorch', 'NER', 'tagger'),
    description='NLP tool',
    long_description='NLP tool, NER, POS',
    license='MIT Licence',
    url='https://github.com/infinity-future/torch-tagger',
    author='infinityfuture',
    author_email='infinityfuture@foxmail.com',
    packages=find_packages(),
    include_package_data=True,
    platforms='any',
    install_requires=INSTALL_REQUIRES
)
